﻿// MinCMS noninvasive ASP.NET WebForms System
// ----------------------------------------------------
// License: Microsoft Reciprocal License (Ms-RL)
// Version: 1.0.0.0
// Initial Development: Prosodev - Software Company / August 2013 / V 1.0.0.0

namespace MinCms.Core
{
    using System;
    using System.Collections;
    using System.Collections.Generic;
    using System.Configuration;
    using System.Data;
    using System.Data.SqlClient;
    using System.Data.SqlServerCe;
    using System.Globalization;
    using System.IO;
    using System.Linq;
    using System.Text;
    using System.Text.RegularExpressions;
    using System.Web;
    using System.Web.Caching;
    using System.Web.Hosting;
    using MinCmsDataSetTableAdapters;
    using Plugins;

    public sealed class MinCms
    {
        #region Constants

        public const string HandledFileTypes = "css|js|gif|jpg|jpeg|png|htm|html|ico|svc|asmx|aspx|ascx|eot|otf|svg|ttf|woff";

        #endregion

        #region Static Fields

        private static MinCms instance;

        #endregion

        #region Fields

        private bool editMode;
        private List<IFile> files;

        #endregion

        #region Constructors and Destructors

        private MinCms()
        {
            this.Properties = new MinCmsProperties("System");
            this.Database = new MinCmsSqlCompactServer();
            this.VirtualPathProvider = new MinCmsPathProvider();
            HostingEnvironment.RegisterVirtualPathProvider(this.VirtualPathProvider);
            this.Plugins = new List<IMinCmsPlugin>
            {
                new MinCmsControl(),
                new LoadControl(),
                new ContentControl(),
                new LinkControl(),
                new TemplateControl(),
                new MinCmsFileEditor(),
                new ActivePageIndicator()
            };
        }

        #endregion

        #region Interfaces

        public interface IDatabase
        {
            #region Public Properties

            IReadOnlyCollection<IFile> Files { get; }

            #endregion

            #region Public Methods and Operators

            void AddFile(IFile file, IFile template, DateTime? insertDate = null, bool? system = null);

            void AddFile(string filePathOrUrl, string templatePath, DateTime? insertDate = null, bool? system = null);

            void ClearDb();

            void DeleteFile(IFile file);

            void DeleteValue(string type, string key);

            void DeleteValue(IFile typeFile, string key);

            IReadOnlyList<string> GetKeys(IFile typeFile);

            IReadOnlyList<string> GetKeys(string type);

            string[] GetLogs(int minLevel, int maxCount = 100);

            IEnumerable<string> GetMinCmsFileUpdateList(DateTime lastAccessTime);

            IEnumerable<string> GetMinCmsPropertyUpdateList(DateTime lastAccessTime);

            int GetMinCmsFileUpdateListCounter(DateTime lastAccessTime);

            int GetMinCmsPropertyUpdateListCounter(DateTime lastAccessTime);

            string GetValue(string type, string key, string defaultValue);

            string GetValue(IFile typeFile, string key, string defaultValue);

            DateTime LastUpdate();

            void Log(int level, string message);

            void MakeRevision();

            void ResetFiles();

            void SetMinCmsUpdateFile(string data);

            void SetValue(string type, string key, string value, bool systemValue = false);

            void SetValue(IFile typeFile, string key, string value, bool systemValue = false);

            #endregion

            void SetMinCmsUpdateProperty(string data);
        }

        public interface IFile
        {
            #region Public Properties

            byte[] Data { get; set; }

            string DataStr { get; set; }

            bool ExistsInDatabase { get; }

            bool ExistsOnDisc { get; }

            Guid HashCode { get; }

            string PhysicalPath { get; }

            string PhysicalRelativePath { get; }

            MinCmsProperties Properties { get; }

            bool SystemFile { get; }

            IFile Template { get; }

            string Url { get; }
            DateTime Version { get; }

            #endregion
        }

        #endregion

        #region Public Properties

        /// <summary>
        ///     Gets the instance to the only MinCms-Class in system (Singleton)
        /// </summary>
        public static MinCms Instance
        {
            get { return instance ?? (instance = new MinCms()); }
        }

        /// <summary>
        ///     Database interface for MinCms. Default is a SQLCE-Provider.
        /// </summary>
        public IDatabase Database { get; set; }

        public IFile DefaultPage
        {
            get
            {
                var startUrl = Instance.Properties["DefaultPage", "~/start.aspx", true];
                var startPage = Instance.Files.FirstOrDefault(a => a.Url == startUrl.ToLower());

                if (startPage == null)
                {
                    return null;
                }

                return startPage;
            }

            set
            {
                if (value == null)
                {
                    Instance.Properties["DefaultPage"] = string.Empty;
                    return;
                }

                Instance.Properties["DefaultPage"] = value.Url;
            }
        }

        /// <summary>
        ///     Gets a value indication if you are in EditMode.
        /// </summary>
        /// <remarks>
        ///     EditMode can be set on localhost only. Web.Config can be used to disable EditMode
        ///     (AppSetting.EditModeDisabled).
        /// </remarks>
        public bool EditMode
        {
            get { return this.LocalhostMode && this.editMode; }
            set
            {
                this.editMode = value;
                this.ResetFiles();
            }
        }

        /// <summary>
        ///     Gets a list of all available files in system.
        /// </summary>
        public IEnumerable<IFile> Files
        {
            get
            {
                lock (Instance)
                {
                    if (this.files == null)
                    {
                        this.files = new List<IFile>();

                        if (!MinCmsHelper.IgnoreFilesOnDisc)
                        {
                            foreach (
                                var fl in
                                    (HandledFileTypes).Split('|')
                                        .Aggregate(string.Empty, (current, k) => current + (";" + "*." + k))
                                        .Split(';')
                                        .ToList())
                            {
                                this.files.AddRange(
                                    Directory.GetFiles(PhysicalRootDirectory, fl, SearchOption.AllDirectories)
                                        .Select(file => new MinCmsFile(DateTime.UtcNow, false, file))
                                        .ToList());
                            }
                        }

                        this.files = this.files.DistinctBy(a => a.PhysicalPath).ToList();

                        if (this.Database != null && this.Database.Files != null)
                        {
                            foreach (var dbFile in this.Database.Files)
                            {
                                if (this.files.FirstOrDefault(a => a.Url == dbFile.Url) == null)
                                {
                                    this.files.Add(dbFile);
                                }
                                else
                                {
                                    this.files.Remove(this.files.First(a => a.Url == dbFile.Url));
                                    this.files.Add(dbFile);
                                }
                            }
                        }

                        this.files =
                            this.files.SortInGroups(a => Path.GetDirectoryName(a.PhysicalPath),
                                a => Path.GetFileName(a.PhysicalPath)).ToList();
                    }
                }

                return this.files;
            }
        }

        public List<IMinCmsPlugin> Plugins { get; private set; }

        public MinCmsProperties Properties { get; private set; }

        public VirtualPathProvider VirtualPathProvider { get; set; }

        #endregion

        #region Properties

        internal static string PhysicalRootDirectory
        {
            get
            {
                var rootDir = HttpContext.Current.Server.MapPath("~");
                if (!rootDir.EndsWith("\\"))
                {
                    rootDir += "\\";
                }

                return rootDir;
            }
        }

        internal bool LocalhostMode
        {
            get { return HttpContext.Current.Request.IsLocal && ConfigurationManager.AppSettings["EditModeDisabled"] != "1"; }
        }

        #endregion

        #region Public Methods and Operators

        /// <summary>
        ///     If a page request is changed (e.g. a modulehandler like MinCmsHtmlCompression), a callback to your
        ///     site will be nesessary. Otherwise the changes don't take effect. The callback is realized with
        ///     default.aspx. The current page will be retrived from HttpContext.Current and default.aspx will be called
        ///     after 2 secounds your page is called again.
        /// </summary>
        /// <remarks>For administrative tasks only.</remarks>
        public void CallBackCurrentPage()
        {
            var urlToCallBack = HttpContext.Current.Request.AppRelativeCurrentExecutionFilePath;
            HttpContext.Current.Session["CallBackCurrentPage"] = urlToCallBack;
            HttpContext.Current.Response.Redirect("~/default.aspx");
        }

        public void ClearMinCmsPluginConfiguration()
        {
            var pg = (string) HttpContext.Current.Session["PluginConfigurationCall_lastDest"];
            HttpContext.Current.Session["PluginConfigurationCall"] = null;
            HttpContext.Current.Session["PluginConfigurationCall_plugin"] = null;
            HttpContext.Current.Session["PluginConfigurationCall_file"] = null;
            HttpContext.Current.Session["PluginConfigurationCall_lastDest"] = null;
            HttpContext.Current.Response.Redirect(pg);
        }

        public void ExecuteMinCmsPluginConfiguration(IMinCmsPlugin plugin)
        {
            var urlToCallBack = HttpContext.Current.Request.AppRelativeCurrentExecutionFilePath;
            HttpContext.Current.Session["PluginConfigurationCall"] = urlToCallBack;
            HttpContext.Current.Session["PluginConfigurationCall_plugin"] = plugin;
            HttpContext.Current.Session["PluginConfigurationCall_file"] = HttpContext.Current.Request.ToIFile();
            HttpContext.Current.Response.Redirect("~/default.aspx");
        }

        /// <summary>
        ///     VirtualPathProvider caches files. If you manually modify files, call this function to publish
        ///     your changes to system.
        /// </summary>
        public void ResetFiles()
        {
            this.files = null;
            this.Database.ResetFiles();
            this.Database.Log(100, "Reset files");
        }

        #endregion

        public sealed class MinCmsProperties
        {
            #region Constructors and Destructors

            public MinCmsProperties()
                : this(string.Empty)
            {
            }

            public MinCmsProperties(string typeVal)
            {
                this.TypeVal = typeVal;
            }

            #endregion

            #region Public Properties

            public int Count
            {
                get { return this.Keys.Count; }
            }

            public IReadOnlyCollection<string> Keys
            {
                get { return Instance.Database.GetKeys(this.TypeVal); }
            }

            public string TypeVal { get; set; }

            #endregion

            #region Public Indexers

            public string this[string key, string defaultValue = "", bool createIfNotExisting = false]
            {
                get { return this.GetValue(key, defaultValue, createIfNotExisting); }

                set { this.SetValue(key, value); }
            }

            #endregion

            #region Public Methods and Operators

            public void Delete(string key)
            {
                Instance.Database.DeleteValue(this.TypeVal, key);
            }

            public string GetValue(string key, string defaultValue, bool createIfNotExisting = false)
            {
                if (createIfNotExisting)
                {
                    if (!Instance.Database.GetKeys(this.TypeVal).Contains(key))
                    {
                        Instance.Database.SetValue(this.TypeVal, key, defaultValue, this.TypeVal == "System");
                    }
                }

                return Instance.Database.GetValue(this.TypeVal, key, defaultValue);
            }

            public bool GetValue(string key, bool defaultValue)
            {
                bool val;
                return bool.TryParse(this.GetValue(key, defaultValue.ToString()), out val) ? val : defaultValue;
            }

            public int GetValue(string key, int defaultValue)
            {
                int val;
                return int.TryParse(this.GetValue(key, defaultValue.ToString(CultureInfo.InvariantCulture)), out val) ? val : defaultValue;
            }

            public double GetValue(string key, double defaultValue)
            {
                double val;
                return double.TryParse(this.GetValue(key, defaultValue.ToString(CultureInfo.InvariantCulture)), out val) ? val : defaultValue;
            }

            public void SetValue(string key, string value)
            {
                Instance.Database.SetValue(this.TypeVal, key, value, this.TypeVal == "System");
            }

            public void SetValue(string key, bool value)
            {
                Instance.Database.SetValue(this.TypeVal, key, value.ToString(), this.TypeVal == "System");
            }

            public void SetValue(string key, int value)
            {
                Instance.Database.SetValue(this.TypeVal, key, value.ToString(CultureInfo.InvariantCulture), this.TypeVal == "System");
            }

            public void SetValue(string key, double value)
            {
                Instance.Database.SetValue(this.TypeVal, key, value.ToString(CultureInfo.InvariantCulture), this.TypeVal == "System");
            }

            #endregion
        }

        internal class Compressor : Stream
        {
            #region Fields

            private readonly Stream stream;

            #endregion

            #region Constructors and Destructors

            public Compressor(Stream stream)
            {
                this.stream = stream;
            }

            #endregion

            #region Public Properties

            public override bool CanRead
            {
                get { return true; }
            }

            public override bool CanSeek
            {
                get { return true; }
            }

            public override bool CanWrite
            {
                get { return true; }
            }

            public override long Length
            {
                get { return 0; }
            }

            public override long Position { get; set; }

            #endregion

            #region Public Methods and Operators

            public override void Close()
            {
                this.stream.Close();
            }

            public override void Flush()
            {
                this.stream.Flush();
            }

            public override int Read(byte[] buffer, int offset, int count)
            {
                return this.stream.Read(buffer, offset, count);
            }

            public override long Seek(long offset, SeekOrigin origin)
            {
                return this.stream.Seek(offset, origin);
            }

            public override void SetLength(long value)
            {
                this.stream.SetLength(value);
            }

            public override void Write(byte[] buffer, int offset, int count)
            {
                var data = new byte[count];
                Buffer.BlockCopy(buffer, offset, data, 0, count);
                var content = Encoding.Default.GetString(buffer);
                if (!Instance.EditMode)
                {
                    content = Regex.Replace(content, "^\\s*", string.Empty, RegexOptions.Compiled | RegexOptions.Multiline);
                    content = Regex.Replace(content, "\\r\\n", string.Empty, RegexOptions.Compiled | RegexOptions.Multiline);
                    content = Regex.Replace(content, "<!--*.*?-->", string.Empty, RegexOptions.Compiled | RegexOptions.Multiline);
                    content = content.Replace("<![CDATA[", "<![CDATA[" + Environment.NewLine);
                    content = content.Replace("//]]>", Environment.NewLine + "//]]>");
                }

                var output = Encoding.Default.GetBytes(content);
                this.stream.Write(output, 0, output.GetLength(0));
            }

            #endregion
        }

        private class MinCmsFile : IFile
        {
            #region Fields

            private readonly GetDataHandler getHandler;

            private readonly SetDataHandler setHandler;

            private readonly string templatePath;

            private IFile template;

            #endregion

            #region Constructors and Destructors

            public MinCmsFile(DateTime version, bool systemFile, string filePathOrUrl, string templatePath = null, GetDataHandler getHandler = null,
                SetDataHandler setHandler = null)
            {
                this.templatePath = templatePath;
                this.getHandler = getHandler;
                this.setHandler = setHandler;
                this.SystemFile = systemFile;
                this.Version = version;

                if (filePathOrUrl.StartsWith("~"))
                {
                    this.Url = filePathOrUrl.ToLower();
                    this.PhysicalPath = this.Url.Replace("~/", PhysicalRootDirectory.ToLower()).Replace("/", "\\");
                    this.PhysicalRelativePath = this.PhysicalPath.Replace(PhysicalRootDirectory.ToLower(), ".\\");
                }
                else
                {
                    this.PhysicalPath = filePathOrUrl.ToLower();
                    this.PhysicalRelativePath = this.PhysicalPath.Replace(PhysicalRootDirectory.ToLower(), ".\\");
                    this.Url = this.PhysicalRelativePath.ToLower().Replace(".\\", "~/").Replace("\\", "/");
                }

                this.HashCode = Guid.NewGuid();
                this.Properties = new MinCmsProperties(this.Url);
            }

            #endregion

            #region Delegates

            public delegate void GetDataHandler(string url, out byte[] data);

            public delegate void SetDataHandler(string url, byte[] data);

            #endregion

            #region Public Properties

            public byte[] Data
            {
                get
                {
                    byte[] dataToReturn = null;

                    if (this.Template != null)
                    {
                        dataToReturn = this.Template.Data;
                        var dt = Encoding.UTF8.GetString(dataToReturn);
                        dt = InjectPreAspNetCode(this, dt);
                        return Encoding.UTF8.GetBytes(dt);
                    }

                    if (this.ExistsOnDisc)
                    {
                        dataToReturn = File.ReadAllBytes(this.PhysicalPath);
                    }

                    if (this.ExistsInDatabase && dataToReturn == null)
                    {
                        if (this.getHandler != null)
                        {
                            this.getHandler(this.Url, out dataToReturn);
                        }
                    }

                    if (!this.Url.EndsWith(".aspx") || dataToReturn == null)
                    {
                        return dataToReturn;
                    }

                    return dataToReturn;
                }

                set
                {
                    if (this.Template != null)
                    {
                        this.Template.Data = value;
                        return;
                    }

                    if (this.ExistsOnDisc)
                    {
                        return;
                    }

                    if (this.ExistsInDatabase)
                    {
                        if (this.setHandler != null)
                        {
                            this.setHandler(this.Url, value);
                        }
                    }

                    Instance.ResetFiles();
                }
            }

            public string DataStr
            {
                get { return this.Template != null ? this.Template.DataStr : Encoding.UTF8.GetString(this.Data); }

                set
                {
                    if (this.Template != null)
                    {
                        this.Template.DataStr = value;
                        return;
                    }

                    this.Data = Encoding.UTF8.GetBytes(value);
                }
            }

            public bool ExistsInDatabase
            {
                get { return Instance.Database.Files.FirstOrDefault(a => a.Url == this.Url) != null; }
            }

            public bool ExistsOnDisc
            {
                get { return File.Exists(this.PhysicalPath) && !MinCmsHelper.IgnoreFilesOnDisc; }
            }

            public Guid HashCode { get; private set; }

            public string PhysicalPath { get; private set; }

            public string PhysicalRelativePath { get; private set; }

            public MinCmsProperties Properties { get; private set; }

            public bool SystemFile { get; private set; }

            public IFile Template
            {
                get
                {
                    if (this.template == null && !string.IsNullOrEmpty(this.templatePath))
                    {
                        this.template = Instance.Files.First(a => a.Url == this.templatePath);
                    }

                    return this.template;
                }
            }

            public string Url { get; private set; }
            public DateTime Version { get; private set; }

            #endregion

            #region Methods

            private static string InjectPreAspNetCode(IFile file, string data)
            {
                var foundTags = true;
                while (foundTags)
                {
                    foundTags = false;
                    foreach (var tag in Instance.Plugins.Select(a => a.GetAdminDisplayName()))
                    {
                        if (tag == "LoadControl")
                        {
                            continue;
                        }

                        var fullTypeName = Instance.Plugins.First(a => a.GetAdminDisplayName() == tag).FullTypeName;
                        var regexNonLoadControls = new Regex("!!" + tag + "([^!]*)!");
                        var nonLoadControlsMatches = regexNonLoadControls.Matches(data);
                        if (nonLoadControlsMatches.Count <= 0)
                        {
                            continue;
                        }

                        var t = Type.GetType(fullTypeName);
                        if (t == null)
                        {
                            continue;
                        }

                        foreach (Match match in nonLoadControlsMatches)
                        {
                            foundTags = true;
                            var s = match.Groups[1].ToString();
                            var functionResult = string.Empty;

                            var m = t.GetInterface("MinCms.Core.IMinCmsPreAspNetRendering");
                            if (m != null)
                            {
                                var obj = Activator.CreateInstance(t);
                                var conf = obj as IMinCmsControlConfiguration;
                                var preAsp = obj as IMinCmsPreAspNetRendering;

                                if (conf != null)
                                {
                                    conf.Configuration = HttpUtility.HtmlDecode(s);
                                }

                                if (preAsp != null)
                                {
                                    functionResult = preAsp.GetPreAspNetRendering(file);
                                }
                            }

                            data = m != null ? data.Replace("!!" + tag + s + "!", functionResult) : data.Replace("!!" + tag + s + "!", "!!LoadControl;" + fullTypeName + s + "!");
                        }
                    }
                }

                var regexLoadControls = new Regex("!!LoadControl([^!]*)!");
                var loadControlsMatches = regexLoadControls.Matches(data);
                if (loadControlsMatches.Count > 0)
                {
                    data = (from Match match in loadControlsMatches select match.Groups[1].ToString()).Aggregate(data,
                        (current, s) =>
                            current.Replace("!!LoadControl" + s + "!",
                                "<mincms:MinCmsControlGenerator runat=\"server\" Configuration=\"" + HttpContext.Current.Server.HtmlEncode(s.StartsWith(";") ? s.Substring(1) : s) + "\" />"));
                }

                return data;
            }

            #endregion
        }

        private class MinCmsPathProvider : VirtualPathProvider
        {
            #region Public Methods and Operators

            public override bool FileExists(string virtualPath)
            {
                return IsVirtualPath(virtualPath) || base.FileExists(virtualPath);
            }

            public override CacheDependency GetCacheDependency(string virtualPath, IEnumerable virtualPathDependencies,
                DateTime utcStart)
            {
                return IsVirtualPath(virtualPath)
                    ? null
                    : base.GetCacheDependency(virtualPath, virtualPathDependencies, utcStart);
            }

            public override VirtualFile GetFile(string virtualPath)
            {
                if (IsVirtualPath(virtualPath))
                {
                    virtualPath = TrimVirtual(virtualPath);
                    var item = Instance.Files.FirstOrDefault(a => a.Url == virtualPath.ToLower());
                    if (item != null)
                    {
                        return this.GetVirtualFile(item);
                    }
                }

                return base.GetFile(virtualPath);
            }

            public override string GetFileHash(string virtualPath, IEnumerable virtualPathDependencies)
            {
                if (IsVirtualPath(virtualPath))
                {
                    virtualPath = TrimVirtual(virtualPath);
                    var item = Instance.Files.FirstOrDefault(a => a.Url == virtualPath.ToLower());
                    if (item != null)
                    {
                        return item.HashCode.ToString();
                    }
                }

                return base.GetFileHash(virtualPath, virtualPathDependencies);
            }

            #endregion

            #region Methods

            private static bool IsVirtualPath(string virtualPath)
            {
                var checkPath = TrimVirtual(virtualPath);
                var result = Instance.Files.FirstOrDefault(a => a.Url == checkPath) != null && !Instance.Files.First(a => a.Url == checkPath).ExistsOnDisc;
                return result;
            }

            private static string TrimVirtual(string virtualPath)
            {
                var url2 = VirtualPathUtility.ToAbsolute("~");
                if (url2.Length == 1)
                {
                    url2 = string.Empty;
                }
                virtualPath = virtualPath.Substring(url2.Length);

                var checkPath = virtualPath.ToLower();
                if (checkPath.StartsWith("/"))
                {
                    checkPath = "~" + checkPath;
                }
                return checkPath;
            }

            private VirtualFile GetVirtualFile(IFile file)
            {
                return new MinCmsVirtualFile(file);
            }

            #endregion

            private class MinCmsVirtualFile : VirtualFile
            {
                #region Fields

                private readonly IFile file;

                #endregion

                #region Constructors and Destructors

                public MinCmsVirtualFile(IFile file)
                    : base(file.Url)
                {
                    this.file = file;
                }

                #endregion

                #region Public Methods and Operators

                public override Stream Open()
                {
                    return new MemoryStream(this.file.Data);
                }

                #endregion
            }
        }

        private class MinCmsSqlCompactServer : IDatabase
        {
            #region Fields

            private List<IFile> files;

            #endregion

            #region Constructors and Destructors

            public MinCmsSqlCompactServer()
            {
                this.CreateDb();
            }

            #endregion

            #region Public Properties

            public IReadOnlyCollection<IFile> Files
            {
                get
                {
                    lock (this)
                    {
                        if (this.files == null)
                        {
                            this.files = new List<IFile>();

                            using (var adapter = new MinCmsFileTableAdapter())
                            {
                                var command = new SqlCeCommand
                                {
                                    Connection = adapter.Connection,
                                    CommandText =
                                        "SELECT Version, Path, Deleted, TemplatePath, System FROM MinCmsFile ORDER BY Version DESC",
                                    CommandType = CommandType.Text
                                };
                                adapter.Adapter.SelectCommand = command;
                                var dt = new MinCmsDataSet.MinCmsFileDataTable();
                                adapter.Adapter.Fill(dt);

                                var rows = dt.DistinctBy(a => a.Path).Where(a => !a.Deleted);
                                this.files =
                                    rows.Select(
                                        a =>
                                            new MinCmsFile(a.Version, !a.IsSystemNull() && a.System, a.Path, !a.IsTemplatePathNull() ? a.TemplatePath : null,
                                                this.GetData, this.SetData) as IFile).ToList();
                                this.files =
                                    this.files.SortInGroups(a => Path.GetDirectoryName(a.PhysicalPath),
                                        a => Path.GetFileName(a.PhysicalPath)).ToList();
                            }
                        }
                    }

                    return this.files;
                }
            }

            #endregion

            #region Public Methods and Operators

            public void AddFile(IFile file, IFile template, DateTime? insertDate = null, bool? system = null)
            {
                lock (this)
                {
                    using (var adapter = new MinCmsFileTableAdapter())
                    {
                        var dt = new MinCmsDataSet.MinCmsFileDataTable();
                        if (template == null)
                        {
                            dt.AddMinCmsFileRow(insertDate.HasValue ? insertDate.Value : DateTime.UtcNow, file.Url, file.Data, false, null, system.HasValue && system.Value);
                        }
                        else
                        {
                            dt.AddMinCmsFileRow(insertDate.HasValue ? insertDate.Value : DateTime.UtcNow, file.Url, null, false, template.Url, system.HasValue && system.Value);
                        }
                        adapter.Update(dt);
                    }
                }

                Instance.ResetFiles();
            }

            public void AddFile(string filePathOrUrl, string templatePath, DateTime? insertDate = null, bool? system = null)
            {
                var file = Instance.Files.FirstOrDefault(a => a.Url == filePathOrUrl.ToLower());
                var template = string.IsNullOrEmpty(templatePath) ? null : Instance.Files.FirstOrDefault(a => a.Url == templatePath.ToLower());

                if (file == null)
                {
                    file = new MinCmsFile(insertDate.HasValue ? insertDate.Value : DateTime.UtcNow, system.HasValue && system.Value, filePathOrUrl, null, this.GetData, this.SetData);
                }

                this.AddFile(file, template, insertDate, system);
            }

            public void ClearDb()
            {
                using (var adapter = new MinCmsFileTableAdapter())
                {
                    var command = new SqlCeCommand
                    {
                        Connection = adapter.Connection,
                        CommandText =
                            "SELECT Version, Path, Deleted, TemplatePath, System FROM MinCmsFile ORDER BY Version DESC",
                        CommandType = CommandType.Text
                    };
                    adapter.Adapter.SelectCommand = command;
                    var dt = new MinCmsDataSet.MinCmsFileDataTable();
                    adapter.Adapter.Fill(dt);
                    foreach (var row in Enumerable.Where(dt, row => row.IsSystemNull() || !row.System))
                    {
                        row.Delete();
                    }
                    adapter.Update(dt);
                }

                using (var adapter = new MinCmsPropertyTableAdapter())
                {
                    var command = new SqlCeCommand
                    {
                        Connection = adapter.Connection,
                        CommandText =
                            "SELECT Version, [Key], [Type], Deleted, System FROM MinCmsProperty ORDER BY Version DESC",
                        CommandType = CommandType.Text
                    };
                    adapter.Adapter.SelectCommand = command;
                    var dt = new MinCmsDataSet.MinCmsPropertyDataTable();
                    adapter.Adapter.Fill(dt);
                    foreach (var row in Enumerable.Where(dt, row => row.IsSystemNull() || !row.System))
                    {
                        row.Delete();
                    }
                    adapter.Update(dt);
                }

                Instance.ResetFiles();
            }

            public void DeleteFile(IFile file)
            {
                lock (this)
                {
                    using (var adapter = new MinCmsFileTableAdapter())
                    {
                        var dt = new MinCmsDataSet.MinCmsFileDataTable();
                        dt.AddMinCmsFileRow(DateTime.UtcNow, file.Url, null, true, null, false);
                        adapter.Update(dt);
                    }
                }

                Instance.ResetFiles();
            }

            public void DeleteValue(string type, string key)
            {
                using (var adapter = new MinCmsPropertyTableAdapter())
                {
                    var dt1 = new MinCmsDataSet.MinCmsPropertyDataTable();
                    dt1.AddMinCmsPropertyRow(DateTime.UtcNow, key, type, null, null, true, false);
                    adapter.Update(dt1);
                }
            }

            public void DeleteValue(IFile typeFile, string key)
            {
                DeleteValue(typeFile.Url, key);
            }

            public IReadOnlyList<string> GetKeys(IFile typeFile)
            {
                return GetKeys(typeFile.Url);
            }

            public IReadOnlyList<string> GetKeys(string type)
            {
                var result = new List<string>();

                using (var adapter = new MinCmsPropertyTableAdapter())
                {
                    var command = new SqlCeCommand
                    {
                        Connection = adapter.Connection,
                        CommandText =
                            "SELECT Version, [Key], [Type], Deleted FROM MinCmsProperty WHERE ([Type] = @param1) ORDER BY Version DESC",
                        CommandType = CommandType.Text
                    };
                    command.Parameters.Add(new SqlCeParameter("@param1", SqlDbType.NVarChar, 500,
                        ParameterDirection.Input, true, 0, 0, "param1", DataRowVersion.Current, null));
                    adapter.Adapter.SelectCommand = command;
                    command.Parameters["@param1"].Value = type;
                    var dt = new MinCmsDataSet.MinCmsPropertyDataTable();
                    adapter.Adapter.Fill(dt);
                    var l = dt.DistinctBy(a => a.Key + a.Type).ToList();
                    result.AddRange(from r in l where !r.Deleted select r.Key);
                    return result;
                }
            }

            public string[] GetLogs(int minLevel, int maxCount = 100)
            {
                using (var adapter = new MinCmsLogTableAdapter())
                {
                    var command = new SqlCeCommand
                    {
                        Connection = adapter.Connection,
                        CommandText =
                            "SELECT TOP(" + maxCount + ") [Version], [Level], [Message] FROM MinCmsLog WHERE ([Level] <= @param1) ORDER BY Version DESC",
                        CommandType = CommandType.Text
                    };
                    command.Parameters.Add(new SqlCeParameter("@param1", SqlDbType.Int));
                    command.Parameters[0].Value = minLevel;

                    var dt = new MinCmsDataSet.MinCmsLogDataTable();
                    adapter.Adapter.SelectCommand = command;
                    adapter.Adapter.Fill(dt);
                    return dt.Select(a => "[" + a.Version.ToShortDateString() + " / " + a.Version.ToLocalTime().ToShortTimeString() + "] " + a.Message).ToArray();
                }
            }

            public IEnumerable<string> GetMinCmsFileUpdateList(DateTime lastAccessTime)
            {
                using (var adapter = new MinCmsFileTableAdapter())
                {
                    var command = new SqlCeCommand
                    {
                        Connection = adapter.Connection,
                        CommandText =
                            "SELECT Version, Path, Deleted, TemplatePath, System FROM MinCmsFile ORDER BY Version DESC",
                        CommandType = CommandType.Text
                    };
                    adapter.Adapter.SelectCommand = command;
                    var dt = new MinCmsDataSet.MinCmsFileDataTable();
                    adapter.Adapter.Fill(dt);

                    foreach (var r in dt.Where(a => a.Version > lastAccessTime))
                    {
                        command = new SqlCeCommand
                        {
                            Connection = adapter.Connection,
                            CommandText =
                                "SELECT Version, Path, Data, Deleted, TemplatePath, System FROM MinCmsFile WHERE Version = @Version AND Path = @Path ORDER BY Version DESC",
                            CommandType = CommandType.Text
                        };
                        command.Parameters.Add(new SqlCeParameter("@Version", SqlDbType.DateTime));
                        command.Parameters.Add(new SqlCeParameter("@Path", SqlDbType.NVarChar));
                        command.Parameters[0].Value = r.Version;
                        command.Parameters[1].Value = r.Path;
                        adapter.Adapter.SelectCommand = command;
                        var dt2 = new MinCmsDataSet.MinCmsFileDataTable();
                        adapter.Adapter.Fill(dt2);

                        var sb = new StringBuilder();
                        var str = new StringWriter(sb);
                        dt2.WriteXml(str);
                        yield return sb.ToString();
                    }
                }
            }

            public IEnumerable<string> GetMinCmsPropertyUpdateList(DateTime lastAccessTime)
            {
                using (var adapter = new MinCmsFileTableAdapter())
                {
                    var command = new SqlCeCommand
                    {
                        Connection = adapter.Connection,
                        CommandText =
                            "SELECT Version, [Key], [Type], Deleted, System FROM MinCmsProperty ORDER BY Version DESC",
                        CommandType = CommandType.Text
                    };
                    adapter.Adapter.SelectCommand = command;
                    var dt = new MinCmsDataSet.MinCmsPropertyDataTable();
                    adapter.Adapter.Fill(dt);

                    foreach (var r in dt.Where(a => a.Version > lastAccessTime))
                    {
                        command = new SqlCeCommand
                        {
                            Connection = adapter.Connection,
                            CommandText =
                                "SELECT Version, [Key], [Type], Value, ValueImage, Deleted, System FROM MinCmsProperty WHERE Version = @Version AND [Key] = @Key AND [Type] = @Type ORDER BY Version DESC",
                            CommandType = CommandType.Text
                        };
                        command.Parameters.Add(new SqlCeParameter("@Version", SqlDbType.DateTime));
                        command.Parameters.Add(new SqlCeParameter("@Key", SqlDbType.NVarChar));
                        command.Parameters.Add(new SqlCeParameter("@Type", SqlDbType.NVarChar));
                        command.Parameters[0].Value = r.Version;
                        command.Parameters[1].Value = r.Key;
                        command.Parameters[2].Value = r.Type;
                        adapter.Adapter.SelectCommand = command;
                        var dt2 = new MinCmsDataSet.MinCmsPropertyDataTable();
                        adapter.Adapter.Fill(dt2);

                        var sb = new StringBuilder();
                        var str = new StringWriter(sb);
                        dt2.WriteXml(str);
                        yield return sb.ToString();
                    }
                }
            }

            public int GetMinCmsFileUpdateListCounter(DateTime lastAccessTime)
            {
                using (var adapter = new MinCmsFileTableAdapter())
                {
                    var command = new SqlCeCommand
                    {
                        Connection = adapter.Connection,
                        CommandText =
                            "SELECT Version, Path, Deleted, TemplatePath, System FROM MinCmsFile ORDER BY Version DESC",
                        CommandType = CommandType.Text
                    };
                    adapter.Adapter.SelectCommand = command;
                    var dt = new MinCmsDataSet.MinCmsFileDataTable();
                    adapter.Adapter.Fill(dt);
                    return dt.Count(a => a.Version > lastAccessTime);
                }
            }

            public int GetMinCmsPropertyUpdateListCounter(DateTime lastAccessTime)
            {
                using (var adapter = new MinCmsPropertyTableAdapter())
                {
                    var command = new SqlCeCommand
                    {
                        Connection = adapter.Connection,
                        CommandText =
                            "SELECT Version, [Key], [Type], Deleted, System FROM MinCmsProperty ORDER BY Version DESC",
                        CommandType = CommandType.Text
                    };
                    adapter.Adapter.SelectCommand = command;
                    var dt = new MinCmsDataSet.MinCmsPropertyDataTable();
                    adapter.Adapter.Fill(dt);
                    return dt.Count(a => a.Version > lastAccessTime);
                }                
            }

            public string GetValue(string type, string key, string defaultValue)
            {
                using (var adapter = new MinCmsPropertyTableAdapter())
                {
                    var command = new SqlCeCommand
                    {
                        Connection = adapter.Connection,
                        CommandText =
                            "SELECT Version, [Key], [Type], Value, ValueImage, Deleted FROM MinCmsProperty WHERE ([Key] = @param2) AND ([Type] = @param1) ORDER BY Version DESC",
                        CommandType = CommandType.Text
                    };
                    command.Parameters.Add(new SqlCeParameter("@param1", SqlDbType.NVarChar, 500,
                        ParameterDirection.Input, true, 0, 0, "param1", DataRowVersion.Current, null));
                    command.Parameters.Add(new SqlCeParameter("@param2", SqlDbType.NVarChar, 500,
                        ParameterDirection.Input, true, 0, 0, "param2", DataRowVersion.Current, null));
                    adapter.Adapter.SelectCommand = command;
                    command.Parameters["@param2"].Value = key;
                    command.Parameters["@param1"].Value = type;
                    var dt = new MinCmsDataSet.MinCmsPropertyDataTable();
                    adapter.Adapter.Fill(dt);
                    if (dt.Count == 0 || dt[0].Deleted)
                    {
                        return defaultValue;
                    }

                    return !dt[0].IsValueImageNull() ? Encoding.UTF8.GetString(dt[0].ValueImage) : dt[0].Value;
                }
            }

            public string GetValue(IFile typeFile, string key, string defaultValue)
            {
                return GetValue(typeFile.Url, key, defaultValue);
            }

            public DateTime LastUpdate()
            {
                var maxDt = DateTime.MinValue;

                foreach (var prop in new List<string> {"MinCmsFile", "MinCmsProperty"})
                {
                    if (prop == "MinCmsFile")
                    {
                        using (var adapter = new MinCmsFileTableAdapter())
                        {
                            var command = new SqlCeCommand
                            {
                                Connection = adapter.Connection,
                                CommandText =
                                    "SELECT Version, Path, Deleted, TemplatePath, System FROM MinCmsFile ORDER BY Version DESC",
                                CommandType = CommandType.Text
                            };
                            adapter.Adapter.SelectCommand = command;
                            var dt = new MinCmsDataSet.MinCmsFileDataTable();
                            adapter.Adapter.Fill(dt);
                            foreach (var r in dt)
                            {
                                if (r.Version > maxDt)
                                {
                                    maxDt = r.Version;
                                }
                            }
                        }
                    }

                    if (prop == "MinCmsProperty")
                    {
                        using (var adapter = new MinCmsPropertyTableAdapter())
                        {
                            var command = new SqlCeCommand
                            {
                                Connection = adapter.Connection,
                                CommandText =
                                    "SELECT Version, [Key], [Type], Deleted FROM MinCmsProperty ORDER BY Version DESC",
                                CommandType = CommandType.Text
                            };
                            adapter.Adapter.SelectCommand = command;
                            var dt = new MinCmsDataSet.MinCmsPropertyDataTable();
                            adapter.Adapter.Fill(dt);
                            foreach (var r in dt)
                            {
                                if (r.Version > maxDt)
                                {
                                    maxDt = r.Version;
                                }
                            }
                        }
                    }
                }

                return maxDt;
            }

            public void Log(int level, string message)
            {
                lock (this)
                {
                    using (var adapter = new MinCmsLogTableAdapter())
                    {
                        var dt = new MinCmsDataSet.MinCmsLogDataTable();
                        dt.AddMinCmsLogRow(DateTime.UtcNow, level, message);
                        adapter.Update(dt);
                    }
                }
            }

            public void MakeRevision()
            {
                using (var adapter = new MinCmsFileTableAdapter())
                {
                    var command = new SqlCeCommand
                    {
                        Connection = adapter.Connection,
                        CommandText =
                            "SELECT Version, Path, Deleted, TemplatePath, System FROM MinCmsFile ORDER BY Version DESC",
                        CommandType = CommandType.Text
                    };
                    adapter.Adapter.SelectCommand = command;
                    var dt = new MinCmsDataSet.MinCmsFileDataTable();
                    adapter.Adapter.Fill(dt);

                    var newestValues = dt.DistinctBy(a => a.Path).ToList();

                    foreach (var nv in Enumerable.Where(dt, nv => !newestValues.Contains(nv)))
                    {
                        if (nv.IsSystemNull() || !nv.System)
                        {
                            nv.Delete();
                        }
                    }

                    adapter.Update(dt);
                }

                using (var adapter = new MinCmsPropertyTableAdapter())
                {
                    var command = new SqlCeCommand
                    {
                        Connection = adapter.Connection,
                        CommandText =
                            "SELECT Version, [Key], [Type], Deleted, System FROM MinCmsProperty ORDER BY Version DESC",
                        CommandType = CommandType.Text
                    };
                    adapter.Adapter.SelectCommand = command;
                    var dt = new MinCmsDataSet.MinCmsPropertyDataTable();
                    adapter.Adapter.Fill(dt);

                    var newestValues = dt.DistinctBy(a => a.Key + a.Type).ToList();

                    foreach (var nv in Enumerable.Where(dt, nv => !newestValues.Contains(nv)))
                    {
                        if (nv.IsSystemNull() || !nv.System)
                        {
                            nv.Delete();
                        }
                    }

                    adapter.Update(dt);
                }

                Compact();
            }

            public void ResetFiles()
            {
                lock (this)
                {
                    this.files = null;
                }
            }

            public void SetMinCmsUpdateFile(string data)
            {
                var dt = new MinCmsDataSet.MinCmsFileDataTable();
                dt.ReadXml(new StringReader(data));
                using (var adapter = new MinCmsFileTableAdapter())
                {
                    var dt2 = new MinCmsDataSet.MinCmsFileDataTable();
                    dt2.ImportRow(dt[0]);
                    adapter.Update(dt2);
                }
            }

            public void SetValue(string type, string key, string value, bool systemValue = false)
            {
                using (var adapter = new MinCmsPropertyTableAdapter())
                {
                    var dt1 = new MinCmsDataSet.MinCmsPropertyDataTable();
                    var bindata = Encoding.UTF8.GetBytes(value);
                    var saveAsBinary = bindata.Length > 450;
                    dt1.AddMinCmsPropertyRow(DateTime.UtcNow, key, type, saveAsBinary ? null : value,
                        saveAsBinary ? bindata : null, false, systemValue);
                    adapter.Update(dt1);
                }
            }

            public void SetValue(IFile typeFile, string key, string value, bool systemValue = false)
            {
                SetValue(typeFile.Url, key, value, systemValue);
            }

            public void SetMinCmsUpdateProperty(string data)
            {
                var dt = new MinCmsDataSet.MinCmsPropertyDataTable();
                dt.ReadXml(new StringReader(data));
                using (var adapter = new MinCmsPropertyTableAdapter())
                {
                    var dt2 = new MinCmsDataSet.MinCmsPropertyDataTable();
                    dt2.ImportRow(dt[0]);
                    adapter.Update(dt2);
                }                
            }

            #endregion

            #region Methods

            private static void Compact()
            {
                string originalConnectionString;
                using (var adapter = new MinCmsFileTableAdapter())
                {
                    originalConnectionString = adapter.Connection.ConnectionString;
                }

                var builder = new SqlConnectionStringBuilder(originalConnectionString);
                var filePath = builder.DataSource.Replace("|DataDirectory|", AppDomain.CurrentDomain.GetData("DataDirectory").ToString());

                if (File.Exists(filePath))
                {
                    using (var en = new SqlCeEngine(originalConnectionString))
                    {
                        en.Shrink();
                    }
                }
            }

            private void CreateDb()
            {
                string connectionString;
                using (var adapter = new MinCmsFileTableAdapter())
                {
                    connectionString = adapter.Connection.ConnectionString;
                }

                var builder = new SqlConnectionStringBuilder(connectionString);
                var filePath = builder.DataSource.Replace("|DataDirectory|", AppDomain.CurrentDomain.GetData("DataDirectory").ToString());

                if (!File.Exists(filePath))
                {
                    var en = new SqlCeEngine(connectionString);
                    en.CreateDatabase();
                }
                else
                {
                    return;
                }

                var tables = "CREATE TABLE [MinCmsProperty] (" + Environment.NewLine +
                             "  [Version] datetime NOT NULL" + Environment.NewLine +
                             ", [Key] nvarchar(500) NOT NULL" + Environment.NewLine +
                             ", [Type] nvarchar(500) NOT NULL" + Environment.NewLine +
                             ", [Value] nvarchar(500) NULL" + Environment.NewLine +
                             ", [ValueImage] image NULL" + Environment.NewLine +
                             ", [Deleted] bit NOT NULL" + Environment.NewLine +
                             ", [System] bit NULL" + Environment.NewLine +
                             ");" + Environment.NewLine +
                             "GO" + Environment.NewLine +
                             "ALTER TABLE [MinCmsProperty] ADD CONSTRAINT [PK_MinCmsProperty] PRIMARY KEY ([Version],[Key],[Type]);" + Environment.NewLine +
                             "GO" +
                             "CREATE TABLE [MinCmsFile] (" + Environment.NewLine +
                             "  [Version] datetime NOT NULL" + Environment.NewLine +
                             ", [Path] nvarchar(500) NOT NULL" + Environment.NewLine +
                             ", [Data] image NULL" + Environment.NewLine +
                             ", [Deleted] bit NOT NULL" + Environment.NewLine +
                             ", [TemplatePath] nvarchar(500) NULL" + Environment.NewLine +
                             ", [System] bit NULL" + Environment.NewLine +
                             ");" + Environment.NewLine +
                             "GO" + Environment.NewLine +
                             "ALTER TABLE [MinCmsFile] ADD CONSTRAINT [PK_MinCmsFile] PRIMARY KEY ([Version],[Path]);" + Environment.NewLine +
                             "GO" + Environment.NewLine +
                             "CREATE TABLE [MinCmsLog] (" + Environment.NewLine +
                             "  [Version] datetime NOT NULL" + Environment.NewLine +
                             ", [Level] int NOT NULL" + Environment.NewLine +
                             ", [Message] nvarchar(500) NOT NULL" + Environment.NewLine +
                             ");" + Environment.NewLine +
                             "GO" + Environment.NewLine +
                             "ALTER TABLE [MinCmsLog] ADD CONSTRAINT [PK_MinCmsLog] PRIMARY KEY ([Version],[Level],[Message]);" + Environment.NewLine +
                             "GO";

                using (var adapter = new MinCmsFileTableAdapter())
                {
                    var con = adapter.Connection;
                    var cmd = con.CreateCommand();
                    var commands = tables.Split(new[] {"GO"}, StringSplitOptions.RemoveEmptyEntries);
                    foreach (var command in commands)
                    {
                        cmd.Connection.Open();
                        cmd.CommandText = command;
                        cmd.ExecuteNonQuery();
                        cmd.Connection.Close();
                    }
                }

                Instance.ResetFiles();
                this.Log(100, "Database created");
            }

            private void GetData(string path, out byte[] data)
            {
                using (var adapter = new MinCmsFileTableAdapter())
                {
                    var command = new SqlCeCommand
                    {
                        Connection = adapter.Connection,
                        CommandText =
                            "SELECT Version, Path, Deleted, Data, TemplatePath FROM MinCmsFile WHERE (Path = @param1) ORDER BY Version DESC",
                        CommandType = CommandType.Text
                    };
                    command.Parameters.Add(new SqlCeParameter("@param1", SqlDbType.NVarChar, 500,
                        ParameterDirection.Input, true, 0, 0, "Type", DataRowVersion.Current, null));
                    command.Parameters["@param1"].Value = path;
                    adapter.Adapter.SelectCommand = command;
                    var dt = new MinCmsDataSet.MinCmsFileDataTable();
                    adapter.Adapter.Fill(dt);
                    if (dt.Count > 0 && !dt[0].IsDataNull())
                    {
                        data = dt[0].Data;
                    }
                    else
                    {
                        data = null;
                    }
                }
            }

            private void SetData(string path, byte[] data)
            {
                lock (this)
                {
                    using (var adapter = new MinCmsFileTableAdapter())
                    {
                        var dt = new MinCmsDataSet.MinCmsFileDataTable();
                        dt.AddMinCmsFileRow(DateTime.UtcNow, path, data, false, null, false);
                        adapter.Update(dt);
                    }
                }

                Instance.ResetFiles();
            }

            #endregion
        }
    }
}