﻿// MinCMS noninvasive ASP.NET WebForms System
// ----------------------------------------------------
// License: Microsoft Reciprocal License (Ms-RL)
// Version: 1.0.0.0
// Initial Development: Prosodev - Software Company / August 2013 / V 1.0.0.0

namespace MinCms.Core
{
    using System;
    using System.Collections.Generic;
    using System.Configuration;
    using System.Data;
    using System.IO;
    using System.Linq;
    using System.Text.RegularExpressions;
    using System.Web;
    using System.Web.UI;
    using System.Xml.Serialization;

    public static class MinCmsHelper
    {
        #region Public Methods and Operators

        public static string DictVal(this Dictionary<string, string> dict, string index)
        {
            var key = dict.Keys.FirstOrDefault(a => String.Equals(a, index, StringComparison.CurrentCultureIgnoreCase));
            if (!string.IsNullOrEmpty(key))
            {
                return dict[key];
            }

            return null;
        }

        public static List<T> FindControls<T>(this Control ctrl)
        {
            var result = new List<T>();
            foreach (var ctrls in ctrl.Controls)
            {
                if (ctrls is T)
                {
                    result.Add((T) ctrls);
                }
                else
                {
                    result.AddRange(((Control) ctrls).FindControls<T>());
                }
            }

            return result;
        }

        public static List<string> GetMinCmsFilesInDirectory(string directoryName)
        {
            if (directoryName.StartsWith("~"))
            {
                directoryName = directoryName.Substring(1);
            }

            directoryName = directoryName.Replace("/", "\\");

            if (directoryName.StartsWith("\\"))
            {
                directoryName = directoryName.Substring(1);
            }

            if (File.Exists(Path.Combine(MinCms.PhysicalRootDirectory, directoryName)))
            {
                return new List<string> {Path.Combine(MinCms.PhysicalRootDirectory, directoryName)};
            }

            var files = new List<string>();
            foreach (var fl in (MinCms.HandledFileTypes).Split('|')
                .Aggregate(string.Empty, (current, k) => current + (";" + "*." + k))
                .Split(';')
                .ToList())
            {
                if (Directory.Exists(Path.Combine(MinCms.PhysicalRootDirectory, directoryName)))
                {
                    files.AddRange(Directory.GetFiles(Path.Combine(MinCms.PhysicalRootDirectory, directoryName), fl, SearchOption.AllDirectories));
                }
            }

            return files;
        }

        public static string AccessToken
        {
            get
            {
                return ConfigurationManager.AppSettings["AccessToken"];
            }
        }

        public static string RemoteSystem
        {
            get
            {
                return ConfigurationManager.AppSettings["RemoteSystem"];
            }
        }

        public static bool IgnoreFilesOnDisc
        {
            get
            {
                return ConfigurationManager.AppSettings["IgnoreFilesOnDisc"] == "True";
            }
        }

        public static bool RemoteSystemHttps { get { return ConfigurationManager.AppSettings["RemoteSystemHttps"] == "True"; } }

        public static MinCms.IFile ToIFile(this HttpRequest request)
        {
            return request == null || string.IsNullOrEmpty(request.AppRelativeCurrentExecutionFilePath)
                ? null
                : MinCms.Instance.Files.FirstOrDefault(a => a.Url == request.AppRelativeCurrentExecutionFilePath.ToLower());
        }

        #endregion

        #region Methods

        internal static MinCmsDataSet.MinCmsFileRow DeserializeXmlToMinCmsFileRow(this string xml)
        {
            if (string.IsNullOrEmpty(xml))
            {
                return null;
            }

            var dt = new MinCmsDataSet.MinCmsFileDataTable();
            dt.ReadXml(new StringReader(xml));
            return dt[0];
        }

        internal static MinCmsDataSet.MinCmsPropertyRow DeserializeXmlToMinCmsPropertyRow(this string xml)
        {
            if (string.IsNullOrEmpty(xml))
            {
                return null;
            }

            var dt = new MinCmsDataSet.MinCmsPropertyDataTable();
            dt.ReadXml(new StringReader(xml));
            return dt[0];
        }

        internal static List<string> DeserializeXmlToStringList(this string xml)
        {
            if (string.IsNullOrEmpty(xml))
            {
                return new List<string>();
            }

            var xmlString = xml;
            var deserializer = new XmlSerializer(typeof (List<string>));
            using (var textReader = new StringReader(xmlString))
            {
                var deserializeCommands = (List<string>) deserializer.Deserialize(textReader);
                textReader.Close();
                return deserializeCommands;
            }
        }

        internal static IEnumerable<TSource> DistinctBy<TSource, TKey>(this IEnumerable<TSource> source, Func<TSource, TKey> keySelector)
        {
            var seenKeys = new HashSet<TKey>();
            return source.Where(element => seenKeys.Add(keySelector(element)));
        }

        internal static string MakeUniqueFilename(this string fileNameToCheck, List<string> listOfFilenames)
        {
            if (listOfFilenames == null)
            {
                return fileNameToCheck;
            }

            var fileNameEnding = Path.GetExtension(fileNameToCheck);

            if (!string.IsNullOrEmpty(fileNameEnding))
            {
                fileNameToCheck = fileNameToCheck.Substring(0, fileNameToCheck.Length - fileNameEnding.Length);
                listOfFilenames = listOfFilenames.Where(a => a.ToLower().EndsWith(fileNameEnding.ToLower())).ToList();
            }

            Func<string, string> generateNextItem = a =>
            {
                var str = a;
                if (str == null)
                {
                    return a;
                }

                var regex = new Regex(@"(\d+)$", RegexOptions.Compiled | RegexOptions.CultureInvariant);
                var match = regex.Match(str);
                var numStr = match.Success ? match.Groups[1].Value : "1";
                if (match.Success)
                {
                    str = str.Substring(0, str.Length - numStr.Length);
                }

                object result = str + (int.Parse(numStr) + 1);
                return (string) result;
            };

            var currentVal = fileNameToCheck;
            var checkVal = true;
            while (checkVal)
            {
                checkVal = false;
                if (!listOfFilenames.Contains(currentVal))
                {
                    continue;
                }

                currentVal = generateNextItem(currentVal);
                checkVal = true;
            }

            return currentVal + fileNameEnding;
        }

        internal static string SerializeToXml(this List<string> list)
        {
            string serializationString;
            var serializer = new XmlSerializer(list.GetType());
            using (var sw = new StringWriter())
            {
                serializer.Serialize(sw, list);
                serializationString = sw.ToString();
            }

            return serializationString;
        }

        internal static string SerializeToXml(this MinCmsDataSet.MinCmsPropertyRow row)
        {
            string serializationString;
            using (var sw = new StringWriter())
            {
                row.Table.WriteXml(sw, XmlWriteMode.IgnoreSchema);
                serializationString = sw.ToString();
            }

            return serializationString;
        }

        internal static string SerializeToXml(this MinCmsDataSet.MinCmsFileRow row)
        {
            string serializationString;
            using (var sw = new StringWriter())
            {
                row.Table.WriteXml(sw, XmlWriteMode.IgnoreSchema);
                serializationString = sw.ToString();
            }

            return serializationString;
        }

        internal static IEnumerable<TSource> SortInGroups<TSource, TKey>(this IEnumerable<TSource> source, Func<TSource, TKey> groupSelector, Func<TSource, TKey> sortSelector)
        {
            var selectGroups = source.GroupBy(groupSelector);
            var resultList = new List<TSource>();
            foreach (var group in selectGroups.OrderBy(a => a.Key))
            {
                resultList.AddRange(@group.OrderBy(sortSelector));
            }

            return resultList;
        }

        internal static string SwitchFileExtension(this string fileName, string ext)
        {
            if (ext.StartsWith("."))
            {
                ext = ext.Substring(1);
            }

            if (fileName.EndsWith("."))
            {
                fileName = fileName.Substring(0, fileName.Length - 1);
            }

            var fname = Path.GetExtension(fileName);
            if (!string.IsNullOrEmpty(fname))
            {
                fname = fileName.Substring(0, fileName.Length - fname.Length);
                if (fname.EndsWith("."))
                {
                    fname = fname.Substring(0, fname.Length - 1);
                }

                fname = fname + "." + ext;
            }
            else
            {
                fname = fileName + "." + ext;
            }

            return fname;
        }

        #endregion
    }
}